#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['batcave']

package_data = \
{'': ['*']}

install_requires = \
['docker ~= 7.1',
 'DotMap ~= 1.3',
 'GitPython ~= 3.1',
 'google-api-core',
 'kubernetes ~= 35.0',
 'requests ~= 2.32',
 'PyYAML ~= 6.0']

extras_require = \
{":platform_machine not in 'aarch64 aarch64_be armv6l armv7l armv8b armv8l'": ['PyQt5 '
                                                                               '~= '
                                                                               '5.15'],
 ":platform_machine not in 'armv6l armv7l armv8b armv8l'": ['psutil ~= 7.2'],
 ":sys_platform == 'win32'": ['pywin32-stubs', 'WMI ~= 1.5'],
 'dev': ['bumpver', 'vjer'],
 'test': ['flake8',
          'flake8-annotations',
          'flake8-pyproject',
          'mypy',
          'pylint',
          'PyQt5-stubs',
          'types-python-dateutil',
          'types-PyYAML',
          'types-psutil',
          'types-requests',
          'unittest-xml-reporting'],
 "test:sys_platform == 'win32'": ['types-pywin32']}

setup(name='batcave',
      version='47.1.0',
      description='Python Programming Toolkit',
      author=None,
      author_email='"Jeffery G. Smith" <web@pobox.com>',
      url=None,
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='~=3.12',
     )
